diff --git a/include/time.h b/include/time.h
index 8ac58e891b..b3e635395d 100644
--- a/include/time.h
+++ b/include/time.h
@@ -25,6 +25,9 @@ libc_hidden_proto (__clock_gettime)
 extern __typeof (clock_settime) __clock_settime;
 libc_hidden_proto (__clock_settime)
 
+extern __typeof (clock_nanosleep) __clock_nanosleep;
+libc_hidden_proto (__clock_nanosleep);
+
 #ifdef __linux__
 extern __typeof (clock_adjtime) __clock_adjtime;
 libc_hidden_proto (__clock_adjtime);
diff --git a/posix/nanosleep.c b/posix/nanosleep.c
index d8564c7119..ed41c8cce7 100644
--- a/posix/nanosleep.c
+++ b/posix/nanosleep.c
@@ -24,10 +24,13 @@ int
 __nanosleep (const struct timespec *requested_time,
 	     struct timespec *remaining)
 {
-  __set_errno (ENOSYS);
-  return -1;
+  int ret = __clock_nanosleep (CLOCK_REALTIME, 0, requested_time, remaining);
+  if (ret != 0)
+    {
+      __set_errno (ret);
+      return -1;
+    }
+  return 0;
 }
-stub_warning (nanosleep)
-
-hidden_def (__nanosleep)
+libc_hidden_def (__nanosleep)
 weak_alias (__nanosleep, nanosleep)
diff --git a/sysdeps/mach/nanosleep.c b/sysdeps/mach/nanosleep.c
deleted file mode 100644
index b60a2179f6..0000000000
--- a/sysdeps/mach/nanosleep.c
+++ /dev/null
@@ -1,79 +0,0 @@
-/* nanosleep -- sleep for a period specified with a struct timespec
-   Copyright (C) 2002-2019 Free Software Foundation, Inc.
-   This file is part of the GNU C Library.
-
-   The GNU C Library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Lesser General Public
-   License as published by the Free Software Foundation; either
-   version 2.1 of the License, or (at your option) any later version.
-
-   The GNU C Library is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <https://www.gnu.org/licenses/>.  */
-
-#include <errno.h>
-#include <mach.h>
-#include <time.h>
-#include <unistd.h>
-
-# define timespec_sub(a, b, result)					      \
-  do {									      \
-    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;			      \
-    (result)->tv_nsec = (a)->tv_nsec - (b)->tv_nsec;			      \
-    if ((result)->tv_nsec < 0) {					      \
-      --(result)->tv_sec;						      \
-      (result)->tv_nsec += 1000000000;					      \
-    }									      \
-  } while (0)
-
-int
-__libc_nanosleep (const struct timespec *requested_time,
-                  struct timespec *remaining)
-{
-  mach_port_t recv;
-  struct timespec before;
-  error_t err;
-
-  if (requested_time->tv_sec < 0
-      || ! valid_nanoseconds (requested_time->tv_nsec))
-    {
-      errno = EINVAL;
-      return -1;
-    }
-
-  const mach_msg_timeout_t ms
-    = requested_time->tv_sec * 1000
-    + (requested_time->tv_nsec + 999999) / 1000000;
-
-  recv = __mach_reply_port ();
-
-  if (remaining != 0)
-    __clock_gettime (CLOCK_REALTIME, &before);
-
-  err = __mach_msg (NULL, MACH_RCV_MSG|MACH_RCV_TIMEOUT|MACH_RCV_INTERRUPT,
-                    0, 0, recv, ms, MACH_PORT_NULL);
-  __mach_port_destroy (mach_task_self (), recv);
-  if (err == EMACH_RCV_INTERRUPTED)
-    {
-      if (remaining != 0)
-	{
-	  struct timespec after, elapsed;
-	  __clock_gettime (CLOCK_REALTIME, &after);
-	  timespec_sub (&after, &before, &elapsed);
-	  timespec_sub (requested_time, &elapsed, remaining);
-	}
-
-      errno = EINTR;
-      return -1;
-    }
-
-  return 0;
-}
-weak_alias(__libc_nanosleep, __nanosleep)
-libc_hidden_def (__nanosleep)
-weak_alias (__libc_nanosleep, nanosleep)
diff --git a/sysdeps/unix/sysv/linux/clock_nanosleep.c b/sysdeps/unix/sysv/linux/clock_nanosleep.c
index 1f240b8720..f3c6fd2d5f 100644
--- a/sysdeps/unix/sysv/linux/clock_nanosleep.c
+++ b/sysdeps/unix/sysv/linux/clock_nanosleep.c
@@ -42,7 +42,7 @@ __clock_nanosleep (clockid_t clock_id, int flags, const struct timespec *req,
   return (INTERNAL_SYSCALL_ERROR_P (r, err)
 	  ? INTERNAL_SYSCALL_ERRNO (r, err) : 0);
 }
-
+libc_hidden_def (__clock_nanosleep)
 versioned_symbol (libc, __clock_nanosleep, clock_nanosleep, GLIBC_2_17);
 /* clock_nanosleep moved to libc in version 2.17;
    old binaries may expect the symbol version it had in librt.  */
diff --git a/sysdeps/unix/sysv/linux/nanosleep.c b/sysdeps/unix/sysv/linux/nanosleep.c
deleted file mode 100644
index 6787909248..0000000000
--- a/sysdeps/unix/sysv/linux/nanosleep.c
+++ /dev/null
@@ -1,31 +0,0 @@
-/* Linux nanosleep syscall implementation.
-   Copyright (C) 2017-2019 Free Software Foundation, Inc.
-   This file is part of the GNU C Library.
-
-   The GNU C Library is free software; you can redistribute it and/or
-   modify it under the terms of the GNU Lesser General Public
-   License as published by the Free Software Foundation; either
-   version 2.1 of the License, or (at your option) any later version.
-
-   The GNU C Library is distributed in the hope that it will be useful,
-   but WITHOUT ANY WARRANTY; without even the implied warranty of
-   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-   Lesser General Public License for more details.
-
-   You should have received a copy of the GNU Lesser General Public
-   License along with the GNU C Library; if not, see
-   <https://www.gnu.org/licenses/>.  */
-
-#include <time.h>
-#include <sysdep-cancel.h>
-#include <not-cancel.h>
-
-/* Pause execution for a number of nanoseconds.  */
-int
-__nanosleep (const struct timespec *requested_time,
-	     struct timespec *remaining)
-{
-  return SYSCALL_CANCEL (nanosleep, requested_time, remaining);
-}
-hidden_def (__nanosleep)
-weak_alias (__nanosleep, nanosleep)
diff --git a/time/clock_nanosleep.c b/time/clock_nanosleep.c
index 5952d29195..d4fc16f6ec 100644
--- a/time/clock_nanosleep.c
+++ b/time/clock_nanosleep.c
@@ -33,7 +33,7 @@ __clock_nanosleep (clockid_t clock_id, int flags, const struct timespec *req,
   /* Not implemented.  */
   return ENOSYS;
 }
-
+libc_hidden_def (__clock_nanosleep)
 versioned_symbol (libc, __clock_nanosleep, clock_nanosleep, GLIBC_2_17);
 /* clock_nanosleep moved to libc in version 2.17;
    old binaries may expect the symbol version it had in librt.  */

diff --git a/sysdeps/unix/clock_nanosleep.c b/sysdeps/mach/clock_nanosleep.c
similarity index 59%
rename from sysdeps/unix/clock_nanosleep.c
rename to sysdeps/mach/clock_nanosleep.c
index 8514a439ee..6a0792d684 100644
--- a/sysdeps/unix/clock_nanosleep.c
+++ b/sysdeps/mach/clock_nanosleep.c
@@ -1,5 +1,5 @@
-/* High-resolution sleep with the specified clock.
-   Copyright (C) 2000-2019 Free Software Foundation, Inc.
+/* clock_nanosleep - high-resolution sleep with specifiable clock.
+   Copyright (C) 2002-2019 Free Software Foundation, Inc.
    This file is part of the GNU C Library.
 
    The GNU C Library is free software; you can redistribute it and/or
@@ -16,28 +16,70 @@
    License along with the GNU C Library; if not, see
    <https://www.gnu.org/licenses/>.  */
 
-#include <assert.h>
 #include <errno.h>
+#include <mach.h>
 #include <time.h>
-#include <sysdep-cancel.h>
+#include <unistd.h>
 #include <shlib-compat.h>
 
-/* This implementation assumes that these is only a `nanosleep' system
-   call.  So we have to remap all other activities.  */
+static struct timespec
+timespec_sub (struct timespec a, struct timespec b)
+{
+  struct timespec result = { a.tv_sec - b.tv_sec,
+			     a.tv_nsec - b.tv_nsec };
+  if (result.tv_nsec < 0)
+    {
+      --result.tv_sec;
+      result.tv_nsec += 1000000000ul;
+    }
+  return result;
+}
+
+static int
+nanosleep_call (const struct timespec *req, struct timespec *rem)
+{
+  mach_port_t recv;
+  struct timespec before;
+  error_t err;
+
+  const mach_msg_timeout_t ms
+    = req->tv_sec * 1000
+    + (req->tv_nsec + 999999) / 1000000;
+
+  recv = __mach_reply_port ();
+
+  if (rem != NULL)
+    __clock_gettime (CLOCK_REALTIME, &before);
+
+  err = __mach_msg (NULL, MACH_RCV_MSG|MACH_RCV_TIMEOUT|MACH_RCV_INTERRUPT,
+                    0, 0, recv, ms, MACH_PORT_NULL);
+  __mach_port_destroy (mach_task_self (), recv);
+  if (err == EMACH_RCV_INTERRUPTED)
+    {
+      if (rem != NULL)
+	{
+	  struct timespec after, elapsed;
+	  __clock_gettime (CLOCK_REALTIME, &after);
+	  elapsed = timespec_sub (after, before);
+	  *rem = timespec_sub (*req, elapsed);
+	}
+
+      return EINTR;
+    }
+
+  return 0;
+}
+
 int
 __clock_nanosleep (clockid_t clock_id, int flags, const struct timespec *req,
 		   struct timespec *rem)
 {
-  struct timespec now;
-
-  if (! valid_nanoseconds (req->tv_nsec))
+  if (clock_id != CLOCK_REALTIME
+      || !valid_nanoseconds (req->tv_nsec)
+      || (flags != 0 || flags != TIMER_ABSTIME))
     return EINVAL;
 
-  if (clock_id == CLOCK_THREAD_CPUTIME_ID)
-    return EINVAL;		/* POSIX specifies EINVAL for this case.  */
-
-  if (clock_id < CLOCK_REALTIME || clock_id > CLOCK_THREAD_CPUTIME_ID)
-    return EINVAL;
+  struct timespec now;
 
   /* If we got an absolute time, remap it.  */
   if (flags == TIMER_ABSTIME)
@@ -68,15 +110,10 @@ __clock_nanosleep (clockid_t clock_id, int flags, const struct timespec *req,
       /* Make sure we are not modifying the struct pointed to by REM.  */
       rem = NULL;
     }
-  else if (flags != 0)
-    return EINVAL;
-  else if (clock_id != CLOCK_REALTIME)
-    /* Not supported.  */
-    return ENOTSUP;
 
-  return __nanosleep (req, rem), 0 ? errno : 0;
+  return nanosleep_call (req, rem);
 }
-
+libc_hidden_def (__clock_nanosleep)
 versioned_symbol (libc, __clock_nanosleep, clock_nanosleep, GLIBC_2_17);
 /* clock_nanosleep moved to libc in version 2.17;
    old binaries may expect the symbol version it had in librt.  */